@@ -1,7 +1,7 @@
 package Net::IMAP::Client;
 
 use vars qw[$VERSION];
-$VERSION = '0.9503';
+$VERSION = '0.9508';
 
 use strict;
 use warnings;
@@ -25,6 +25,7 @@ my %DEFAULT_ARGS = (
     pass     => undef,
     ssl      => 0,
     ssl_verify_peer => 1,
+    tls      => 0,
     socket   => undef,
     _cmd_id  => 0,
     ssl_options => {},
@@ -37,17 +38,21 @@ sub new {
         $_ => exists $args{$_} ? $args{$_} : $DEFAULT_ARGS{$_}
     } keys %DEFAULT_ARGS };
 
+    die "Cannot enable both ssl and tls" if ($self->{tls} and $self->{ssl});
+
     bless $self, $class;
 
     $self->{notifications} = [];
     eval {
-        $self->{greeting} = $self->_socket_getline;
+        $self->_get_socket;     # set up the socket
     };
+
     return $@ ? undef : $self;
 }
 
 sub DESTROY {
     my ($self) = @_;
+	local $@;
     eval {
         $self->quit
           if $self->{socket}->opened;
@@ -279,11 +284,26 @@ sub seq_to_uid {
     return undef;
 }
 
+{ my %no_args = map{ $_, 1} qw(
+	ALL
+	ANSWERED
+	DELETED
+	DRAFT
+	FLAGGED
+	NEW
+	OLD
+	RECENT
+	SEEN
+	UNANSWERED
+	UNDELETED
+	UNDRAFT
+	UNFLAGGED
+	UNSEEN
+	);
 sub search {
     my ($self, $criteria, $sort, $charset) = @_;
 
     $charset ||= 'UTF-8';
-    _string_quote($charset);
 
     my $cmd = $sort ? 'SORT' : 'SEARCH';
     if ($sort) {
@@ -302,10 +322,13 @@ sub search {
     if (ref($criteria) eq 'HASH') {
         my @a;
         while (my ($key, $val) = each %$criteria) {
+            $key = uc $key;
+            push @a, $key;
+            next if $no_args{$key};
+            # don't quote range
             my $quoted = $val;
-			# don't quote range
-			_string_quote($quoted) unless uc $key eq 'UID';
-            push @a, uc $key, $quoted;
+            _string_quote($quoted) unless $key eq 'UID';
+            push @a, $quoted;
         }
         $criteria = '(' . join(' ', @a) . ')';
     }
@@ -324,6 +347,7 @@ sub search {
 
     return undef;
 }
+}
 
 sub get_rfc822_body {
     my ($self, $msg) = @_;
@@ -596,6 +620,7 @@ sub expunge {
 
 sub last_error {
     my ($self) = @_;
+	defined $self->{_error} or return;
     $self->{_error} =~ s/\s+$//s; # remove trailing carriage return
     return $self->{_error};
 }
@@ -636,7 +661,11 @@ sub _get_ssl_config {
     my %ssl_config = ( SSL_verify_mode => IO::Socket::SSL::SSL_VERIFY_PEER );
 
     if ($^O eq 'linux' && !$self->{ssl_ca_path} && !$self->{ssl_ca_file}) {
-        $ssl_config{SSL_ca_path} = '/etc/ssl/certs/';
+        $ssl_config{SSL_ca_path} = 
+			-d '/etc/ssl/certs/' ? '/etc/ssl/certs/' : '/etc/pki/tls/certs/'; 
+
+		-d $ssl_config{SSL_ca_path} 
+			or die "$ssl_config{SSL_ca_path}: SSL certification directory not found";
     }
     $ssl_config{SSL_ca_path} = $self->{ssl_ca_path} if $self->{ssl_ca_path};
     $ssl_config{SSL_ca_file} = $self->{ssl_ca_file} if $self->{ssl_ca_file};
@@ -645,7 +674,11 @@ sub _get_ssl_config {
 }
 sub _get_socket {
     my ($self) = @_;
-     my $socket = $self->{socket} ||= ($self->{ssl} ? 'IO::Socket::SSL' : 'IO::Socket::INET')->new(
+
+    my $socket = $self->{socket};
+    return $socket if (defined($socket) and ($socket->isa('IO::Socket::SSL')or $socket->isa('IO::Socket::INET')));
+
+    $self->{socket} = ($self->{ssl} ? 'IO::Socket::SSL' : 'IO::Socket::INET')->new(
 			( ( %{$self->{ssl_options}} ) x !!$self->{ssl} ), 
                 PeerAddr => $self->_get_server,
                 PeerPort => $self->_get_port,
@@ -654,8 +687,44 @@ sub _get_socket {
                 Blocking => 1,
                 $self->_get_ssl_config,
             ) or die "failed connect or ssl handshake: $!,$IO::Socket::SSL::SSL_ERROR";
-    $socket->sockopt(SO_KEEPALIVE, 1);
-    return $socket;
+    $self->{socket}->sockopt(SO_KEEPALIVE, 1);
+
+    $self->{greeting} = $self->_socket_getline; # get the initial greeting
+
+    $self->_starttls if ($self->{tls});         # upgrade to TLS if needed
+
+    return $self->{socket};
+}
+
+sub _starttls {
+    my ($self) = @_;
+
+    # ask for the capabilities directly at this level, make sure we can do STARTTLS
+    my $can_do_starttls = 0;
+    my ($ok, $lines) = $self->_tell_imap('CAPABILITY');
+    if ($ok) {
+        my $line = $lines->[0][0];
+        $can_do_starttls ||= 1 if ($line =~ /^\*\s+CAPABILITY.*\s+STARTTLS/);
+    } else {
+        die "IMAP server failed CAPABILITY query"
+    }
+    die "IMAP server does not have STARTTLS capability" unless ($can_do_starttls);
+
+    # request STARTTLS
+    ($ok, $lines) = $self->_tell_imap('STARTTLS');
+    if ($ok) {
+        my @sni_args;
+        push(@sni_args, SSL_hostname => $self->_get_server) if (IO::Socket::SSL->can_client_sni());
+        IO::Socket::SSL->start_SSL(
+            $self->{socket},
+            $self->_get_ssl_config,
+            @sni_args,
+        ) or die $IO::Socket::SSL::SSL_ERROR;
+    } else {
+        die "IMAP server failed STARTTLS command"
+    }
+
+    return $self->{socket};
 }
 
 sub _get_next_id {
@@ -982,7 +1051,7 @@ sub _parse_tokens {
                 push @{$stack[-1]}, $str; # found string
             } elsif ($text =~ m/\G(\d+)/gc) {
                 push @{$stack[-1]}, $1 + 0; # found numeric
-            } elsif ($text =~ m/\G([a-zA-Z0-9_\$\\.+\/*&-]+)/gc) {
+            } elsif ($text =~ m/\G ([ a-z A-Z 0-9 _ $ . + \/ & ! # ' , : ; < = > ? @ ^ ` ~ | } - ]+ )/gcxx){
                 my $atom = $1;
                 if (lc $atom eq 'nil') {
                     $atom = undef;
@@ -1198,7 +1267,16 @@ Password
 
 =item - B<ssl> (BOOL, optional, default FALSE)
 
-Pass a true value if you want to use IO::Socket::SSL
+Pass a true value if you want to use L<IO::Socket::SSL>
+You may not set both C<ssl> and C<tls> at the same time.
+
+=item - B<tls> (BOOL, optional, default FALSE)
+
+Pass a true value if you want to use connect without SSL and then use
+C<STARTTLS> to upgrade the connection to an encrypted session using
+L<IO::Socket::SSL>.  The other C<ssl_*> options also apply.
+
+You may not set both C<ssl> and C<tls> at the same time.
 
 =item - B<ssl_verify_peer> (BOOL, optional, default TRUE)
 
